/*
 * Newt-Puzzle
 * Copyright (C) Aude Jaulent 2008, 2009 <elderme@tuxfamily.org>
 * 
 * Newt-Puzzle is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Newt-Puzzle is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "camera.h"


Camera::Camera(float dist, float fov, float nearClip, float farClip, int xWin, int yWin, int width, int height):
mFov(fov), mNearClip(nearClip), mFarClip(farClip), xWindow(xWin), yWindow(yWin), mWidth(width), mHeight(height){
	mPosition = Vector3(0., 0., dist);
	xAxis = Vector3(-1.,0.,0.);
	yAxis = Vector3(0.,1.,0.);
	zAxis = Vector3(0.,0.,-1.);

}

Camera::Camera(Vector3 position, Vector3 target, Vector3 up, float fov, float nearClip, float farClip, int xWin, int yWin, int width, int height) :
	mFov(fov), mNearClip(nearClip), mFarClip(farClip), xWindow(xWin), yWindow(yWin), mWidth(width), mHeight(height){
	setPositionAndAxis (mPosition, target, up);
}

void Camera::setPositionAndAxis(Vector3 position, Vector3 target, Vector3 up) {
	mPosition = position;
	zAxis = target - position ;
	zAxis.normalize();
	yAxis = up - (up*zAxis)*zAxis;
	yAxis.normalize();
	xAxis = yAxis^zAxis;		
}

void Camera::setFarClip(float farClip) {
	mFarClip = farClip;
}

float Camera::getNearClip() {
	return mNearClip;
}

float Camera::getFarClip() {
	return mFarClip;
}


Vector3 Camera::getPosition() {
	return mPosition;
}

Vector3 Camera::getDirection() {
	return zAxis;
}

float Camera::getVAngle() {
	return mFov;
}

float Camera::getHAngle() {
	return mFov* (float) mWidth / (float) mHeight;
}

Vector3 Camera::getTranslatedPosition(Vector3 tr, Vector3 vertical) {
	Vector3 U,W;
	float p = zAxis*vertical;
	float coeff1 = sqrt(1./(1.-p*p));
	float coeff2 = -coeff1*p;
	W = coeff1*zAxis + coeff2*vertical;
	U = vertical^W;
	return (mPosition + tr.x * U + tr.y * vertical + tr.z * W);
}

Vector3 Camera::getXAxis() {
	return xAxis;
}

Vector3 Camera::getYAxis() {
	return yAxis;
}

Vector3 Camera::getZAxis() {
	return zAxis;
}


void Camera::getDisplayParam(int &xWin, int &yWin, int &w, int &h) {
	xWin = xWindow;
	yWin = yWindow;
	w = mWidth;
	h = mHeight;
}

int Camera::getLeft() {
	return xWindow;
}

int Camera::getRight() {
	return xWindow + mWidth;
}

int Camera::getBottom() {
	return yWindow;
}

int Camera::getTop() {
	return yWindow + mHeight;
}


void Camera::perspective() {
	 gluPerspective(mFov, (float)mWidth/(float)mHeight, mNearClip, mFarClip);
}

void Camera::place() {
	Vector3 target = mPosition + zAxis;
	gluLookAt(mPosition.x, mPosition.y, mPosition.z,
			  target.x, target.y, target.z,
			  yAxis.x, yAxis.y, yAxis.z);	
}

void Camera::lookAtTheSky() {
	gluLookAt(0., 0., 0.,
			  zAxis.x, zAxis.y, zAxis.z,
			  yAxis.x, yAxis.y, yAxis.z);	
}


void Camera::initModelview() {
	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void Camera::initProjection() {
	glViewport(xWindow,yWindow,mWidth,mHeight);	
	
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	perspective ();
}

void Camera::initPickingProjection(int xMouse, int yMouse, float delX, float delY) {
	glViewport(xWindow,yWindow,mWidth,mHeight);	
	
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	
	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	gluPickMatrix((GLdouble) xMouse, (GLdouble) (viewport[3] - yMouse), delX, delY, viewport);
}


void Camera::setPosition(Vector3 position) {
	mPosition = position;
}

void Camera::translate(Vector3 tr) {
	mPosition = mPosition + tr.x * xAxis + tr.y * yAxis + tr.z * zAxis ;
}

void Camera::translate(Vector3 tr, Vector3 vertical) {
	Vector3 U,W;
	float p = zAxis*vertical;
	float coeff1 = sqrt(1./(1.-p*p));
	float coeff2 = -coeff1*p;
	W = coeff1*zAxis + coeff2*vertical;
	U = vertical^W;
	mPosition = mPosition + tr.x * U + tr.y * vertical + tr.z * W;
}

void Camera::rotate(float phi, float theta) {

	float cosPhi = cos(phi);
	float sinPhi = sin(phi);
	float cosTheta = cos(theta);
	float sinTheta = sin(theta);
	bool allowedRot = false;
	
	
	//rotation angle phi
	zAxis.x = cosPhi * zAxis.x + sinPhi * zAxis.z;
	zAxis.z = cosPhi * zAxis.z - sinPhi * zAxis.x;
	
	yAxis.x = cosPhi * yAxis.x + sinPhi * yAxis.z;
	yAxis.z = cosPhi * yAxis.z - sinPhi * yAxis.x;	
	
	
	//rotation angle theta
	if(zAxis.y > 0.999) {
		if(theta < 0.) allowedRot = true;
	} else {
		if(zAxis.y < -0.999) {
			if(theta > 0.) allowedRot = true;
		} else {
			
			Vector3 zAxisNew = cosTheta * zAxis + sinTheta * yAxis;
			if((zAxisNew.x * zAxis.x < 0.)||(zAxisNew.z * zAxis.z < 0.)) {
				if(theta > 0.) { //we look at the sky
					zAxis = Vector3(0.,1.,0.);
					float coeff = 1./sqrt(1. - yAxis.y*yAxis.y);
					yAxis = Vector3(coeff * yAxis.x, 0., coeff * yAxis.z);
					
				} else { //we look at the floor
					zAxis = Vector3(0.,-1.,0.);
					float coeff = -1./sqrt(1. - yAxis.y*yAxis.y);
					yAxis = Vector3(coeff * yAxis.x, 0., coeff * yAxis.z);
				}	
					
			} else {
				allowedRot = true;	
			}
		}
	}
	
	if(allowedRot) {
		zAxis = cosTheta * zAxis + sinTheta * yAxis;
		float coeff2 = 1./sqrt(1. - zAxis.y*zAxis.y);
		float coeff1 = -1.*zAxis.y*coeff2;
		yAxis = coeff1 * zAxis + coeff2 * Vector3(0.,1.,0.);
	}

	yAxis.normalize();
	zAxis.normalize();
	xAxis = zAxis ^ yAxis;	
	
}

void Camera::rotate(float phi, float theta, Vector3 vertical) {
	float cosPhi = cos(phi);
	float sinPhi = sin(phi);
	float cosTheta = cos(theta);
	float sinTheta = sin(theta);
	bool allowedRot = false;
	
	//rotation angle phi
	Vector3 W = (zAxis*vertical)*vertical;
	Vector3 U = zAxis - W;
	Vector3 V = vertical^U;
	zAxis = cosPhi * U + sinPhi * V + W;
	
	U = yAxis - W;
	V = vertical^U;
	yAxis = cosPhi * U + sinPhi * V + W;
	
	
	//rotation angle theta
	Vector3 vectRot = zAxis^vertical;
	vectRot.normalize();
	U = vectRot.perpendicular();
	V = vectRot^U; 
			
	Vector3 zAxisNew = cosTheta * zAxis + sinTheta * ((zAxis*U)*V - (zAxis*V)*U);
	
	if(zAxis*vertical > 0.999) {
		if(theta < 0.) allowedRot = true;
	} else {
		if(zAxis*vertical < -0.999) {
			if(theta > 0.) allowedRot = true;
		} else {
					 
			if(((zAxisNew*U) * (zAxis*U) < 0.)||((zAxisNew*V) * (zAxis*V) < 0.)) {
				if(theta > 0.) { //we look at the sky
					zAxis = vertical;
					float dotProd = vertical*yAxis;
					float coeff = 1./sqrt(1. - dotProd * dotProd);
					yAxis = coeff * yAxis - (dotProd * coeff) * vertical;
					
				} else { //we look at the floor
					zAxis = -1.*vertical;
					float dotProd = vertical*yAxis;
					float coeff = - 1./sqrt(1. - dotProd * dotProd);
					yAxis = coeff * yAxis - (dotProd * coeff) * vertical;
				}	
					
			} else {
				allowedRot = true;	
			}
		}
	}
	if(allowedRot) {
		zAxis = zAxisNew;
		float dotProd = vertical*zAxis;
		float coeff2 = 1./sqrt(1. - dotProd*dotProd);
		float coeff1 = -1.*dotProd*coeff2;
		yAxis = coeff1 * zAxis + coeff2 * vertical;
	}

	yAxis.normalize();
	zAxis.normalize();
	xAxis = zAxis ^ yAxis;		
}

void Camera::keplerRotate(float hAngle, float vAngle) {
	
	
	
	
}

void Camera::rotAroundO(bool absoluteYAxis, float dist, float hAngle, float vAngle, float limitUp, float limitDown) {
	float cosh = cos(hAngle);
	float sinh = sin(hAngle);
	float cosv = cos(vAngle);
	float sinv = sin(vAngle);
		
	if(absoluteYAxis) {
		Vector3 Y = Vector3(0.,1.,0.);
		Vector3 zNew;
		
		zNew = cosv * zAxis - sinv * yAxis;
		float yy = yAxis.y;
		
		if(zNew.y > limitDown){
			float coeff1 = (1.-limitDown*limitDown)/(1.-yy*yy);
			coeff1 = -1.*sqrt(coeff1);
			float coeff2 = limitDown - coeff1*yy;
			zNew = coeff1*yAxis + coeff2*Y;
			
		}
		if(zNew.y < -1.*limitUp){
			float coeff1 = (1.-limitUp*limitUp)/(1.-yy*yy);
			coeff1 = sqrt(coeff1);
			float coeff2 = -1.*limitUp - coeff1*yy;
			zNew = coeff1*yAxis + coeff2*Y;		
		}	
		
		zNew.normalize(); 
		zAxis = zNew.rotate(Y, hAngle);
		setzAxisLookingAtO(zAxis);
		
	} else {
		zAxis = cosv * zAxis - sinv * yAxis;
		xAxis = zAxis^yAxis;
		zAxis = cosh*zAxis - sinh*xAxis;		
		yAxis = zAxis^xAxis;
		
		xAxis.normalize();
		yAxis.normalize();
		zAxis.normalize();
	}
		
	mPosition.x = -1.*dist*zAxis.x;
	mPosition.y = -1.*dist*zAxis.y;
	mPosition.z = -1.*dist*zAxis.z;	
}

void Camera::setPositionLookingAtO(Vector3 position) {
	mPosition = position;
	zAxis = -1 * mPosition;
	zAxis.normalize();
	
	Vector3 Y = Vector3(0., 1., 0.);
	float coeff = 1. / sqrt(1.-zAxis.y*zAxis.z);
	yAxis = coeff * Y - (coeff*zAxis.y) * zAxis;
	yAxis.normalize();
	
	xAxis = yAxis^zAxis;
}

void Camera::setzAxisLookingAtO(Vector3 z) {
	Vector3 Y = Vector3(0.,1.,0.);
	zAxis = z;
	
	float coeff = 1. / sqrt(1.-zAxis.y*zAxis.y);
	yAxis = coeff * Y - (coeff*zAxis.y) * zAxis;
	yAxis.normalize();

	xAxis = yAxis^zAxis;	
	xAxis.normalize();
}

void Camera::setDistToO(float dist) {
	mPosition = -dist * zAxis;
}
