/*
 * Newt-Puzzle
 * Copyright (C) Aude Jaulent 2009 <elderme@tuxfamily.org>
 * 
 * Newt-Puzzle is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Newt-Puzzle is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "score.h"
#include "./newtEngine/scene.h"



Score::Score(std::vector<int> colors, int scoreMax, 
			 float baseRadius, int nbRbase, float baseHeight, 
		  	float cylinderRadius, int nbRcyl,
		  	float torusRadius, int nbTorus1, int nbTorus2,
		  	Vector3 position) :
EntityWithPosition(position, Vector3(0.,0.,1.), 90.), nbPlayers(colors.size()-1) {
	
	Material *material;
	material = new Material("materialScoreBase", colors[0]);
	
	float torusSmallRadius = torusRadius - cylinderRadius;
	float interval = 2.2 * torusSmallRadius;
	float cylinderHeight = ((float) scoreMax + 3) * interval;
	
	Entity *base = new EntityWithPosition(material,
										  new CylinderMesh(baseHeight/3., 1, baseRadius, nbRbase, 
														   BASE_FLAT, BASE_FLAT, 4, baseHeight/3., false),
										  true,
										  Vector3(-cylinderHeight/2., 0., 0.),
										  Vector3(0.,1.,0.),0.);
	addChild (base, 0);
	
	CylinderMesh *cylinder = new CylinderMesh(cylinderHeight, 1, cylinderRadius, nbRcyl, 
										     BASE_NONE, BASE_FLAT, nbRcyl/2, cylinderRadius/3., false);
	
	int nbPlayers = colors.size() - 1;
	for(int i=1;i<=nbPlayers;i++) {
		material = new Material("materialScore", colors[i]);
		addChild(new PlayerScore(material, cylinder, baseRadius, cylinderHeight, interval, i, nbPlayers), i);
	}
	
	torusMesh = new TorusMesh(torusRadius, nbTorus1, torusSmallRadius, nbTorus2, false);
	scene->addMeshInTheSet(torusMesh, material->getNumberOfTextures());	
	
}

Score::~Score() {
}

void Score::incScore(int player, int delta) {
	PlayerScore *playerScore = dynamic_cast <PlayerScore*> (children[player]);
	playerScore->incScore(delta, torusMesh);
}

int Score::getScore(int player) {
	PlayerScore *playerScore = dynamic_cast <PlayerScore*> (children[player]);
	return playerScore->getValue();
}

void Score::reinit() {
	for(int i=1;i<=nbPlayers;i++) {
		PlayerScore *playerScore = dynamic_cast <PlayerScore*> (children[i]);
		playerScore->reinit();
	}
}

/************************/
/*	PlayerScore			*/ 
/************************/


PlayerScore::PlayerScore(Material *material, CylinderMesh *cylinder, float baseRadius, 
						 float cylinderHeight, float interval, int position, int nbPlayers) :
EntityWithValue(material, cylinder, 0, true) {
					
	float angle = ((float) position - 1)/ (float) nbPlayers * 2. * M_PI + M_PI /2.;	
	mPosition = Vector3(0., baseRadius/7.*3. * sin(angle), baseRadius/7.*3. * cos(angle));

	lineTorus = new EntityLine(interval, true, Vector3(0., 1., 0.),
									  Vector3((interval - cylinderHeight) / 2., 0.,0.),Vector3(0.,0.,1.), -90.);
	addChild(lineTorus, 0);
	
}
				
				
PlayerScore::~PlayerScore() {
	
}
			
void PlayerScore::applyTransformation () {
	glTranslatef (mPosition.x, mPosition.y, mPosition.z); 
}

void PlayerScore::incScore(int delta, TorusMesh *torusMesh) {
	
	for(int i=0;i<delta;i++) {
		lineTorus->insertChild(NULL, torusMesh, 0, mValue+i);
	}
	mValue += delta;
}

void PlayerScore::reinit() {
	mValue = 0;
	lineTorus->removeAllChildren();
}


