/*
 * Newt-Puzzle
 * Copyright (C) Aude Jaulent 2008, 2009 <elderme@tuxfamily.org>
 * 
 * Newt-Puzzle is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Newt-Puzzle is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "cylinder.h"
#include "../newtEngine/scene.h"


CylinderMesh::CylinderMesh(float height, int nbH, float radius, int nbR, bool inside) :
	Mesh("cylinder", false, true, false, false), mHeight(height), mNbH(nbH), mRadius(radius), mInside(inside),
	mNbRbase(0), mBase1(BASE_NONE), mBase2(BASE_NONE) {
		
	float meshLevel = scene->getGraphicOptions()->getMeshLevel();
	float value;
		
	value = (float) nbR * meshLevel;
	if(value < 4) {
		mNbR = 4;
	} else {
		mNbR = (int) value;
	}		
}

CylinderMesh::CylinderMesh(float height, int nbH, float radius, int nbR, baseType base1, baseType base2, int nbRbase, bool inside) :
	Mesh("cylinder", false, true, false, true), mHeight(height), mNbH(nbH), mRadius(radius), mInside(inside),
	mNbRbase(nbRbase), mBase1(base1), mBase2(base2) {

	float meshLevel = scene->getGraphicOptions()->getMeshLevel();
	float value;
		
	value = (float) nbR * meshLevel;
	if(value < 4) {
		mNbR = 4;
	} else {
		mNbR = (int) value;
	}			
		
	if((mBase1 == BASE_FLAT)||(mBase2 == BASE_FLAT)) {
		mEdgeRadius = mRadius/3.;
	}
}

CylinderMesh::CylinderMesh(float height, int nbH, float radius, int nbR, baseType base1, baseType base2, int nbRbase, float edgeRadius, bool inside):
	Mesh("cylinder", false, true, false, true), mHeight(height), mNbH(nbH), mRadius(radius), mInside(inside),
	mNbRbase(nbRbase), mBase1(base1), mBase2(base2), mEdgeRadius(edgeRadius) {
		
	float meshLevel = scene->getGraphicOptions()->getMeshLevel();
	float value;
		
	value = (float) nbR * meshLevel;
	if(value < 4) {
		mNbR = 4;
	} else {
		mNbR = (int) value;
	}			
}

CylinderMesh::~CylinderMesh() {
}

void CylinderMesh::initFirstParams() {
	geometry.nbVertices = (mNbH + 1) * (mNbR + 1);
	
	geometry.nbIndexArrays = mNbR;
	geometry.primType = new GLenum[geometry.nbIndexArrays];
	for(int i=0;i<geometry.nbIndexArrays;i++) {
		geometry.primType[i] = GL_QUAD_STRIP;
	}
	
	geometry.nbIndices = new int[geometry.nbIndexArrays];
	for(int i=0;i<geometry.nbIndexArrays;i++) {
		geometry.nbIndices[i] = 2 * (mNbH+1);
	}		
	
	if((mBase1 == BASE_NONE)&&(mBase2 == BASE_NONE)) {
		mpUniformTangent = new Vector3(1., 0., 0.);	
	} else {
		switch(mBase1) {
			case BASE_DOME:
				geometry.nbVertices += mNbRbase * (mNbR + 1);
				for(int i=0;i<geometry.nbIndexArrays;i++) {
					geometry.nbIndices[i] += 2 * mNbRbase;
				}			
				break;
			
			case BASE_FLAT:
				geometry.nbVertices += (mNbRbase + 1)* (mNbR + 1);
				for(int i=0;i<geometry.nbIndexArrays;i++) {
					geometry.nbIndices[i] += 2 * (mNbRbase + 1);
				}			
				break;
			
		}
		
		switch(mBase2) {
			case BASE_DOME:
				geometry.nbVertices += mNbRbase * (mNbR + 1);			
				for(int i=0;i<geometry.nbIndexArrays;i++) {
					geometry.nbIndices[i] += 2 * mNbRbase;
				}			
				break;
			
			
			case BASE_FLAT:
				geometry.nbVertices += (mNbRbase + 1)* (mNbR + 1);
				for(int i=0;i<geometry.nbIndexArrays;i++) {
					geometry.nbIndices[i] += 2 * (mNbRbase + 1);
				}				
				break;
			
		}		
	}
}
		
void CylinderMesh::createVertices(Vector3 *ptr) {
	float alpha;
	float lambda;
	
	if(mBase1 == BASE_FLAT) {
		*ptr = Vector3(- 0.5 * mHeight - mEdgeRadius, 0., 0.);
		ptr++;
	}
	
	for(int j=-mNbRbase;j<0;j++) {			
		switch(mBase1) {
			case BASE_DOME:
				lambda = (float) j/(float) mNbRbase* M_PI / 2.;
				*ptr = Vector3(mRadius * sin(lambda) - 0.5 * mHeight, 0., -mRadius * cos(lambda));
				ptr ++;		
				break;
			
			case BASE_FLAT:
				lambda = (float) j/(float) mNbRbase* M_PI / 2.;
				*ptr = Vector3(mEdgeRadius * sin(lambda) - 0.5 * mHeight, 0., -(mRadius - mEdgeRadius) - mEdgeRadius * cos(lambda));
				ptr ++;		
				break;
		}	
	}
	for(int j=0;j<=mNbH;j++) {		
		*ptr = Vector3(((float) (j/mNbH) - 1./2.) * mHeight, 0., -mRadius);
		ptr ++;
	}	
	for(int j=1;j<=mNbRbase;j++) {			
		switch(mBase2) {
			case BASE_DOME:
				lambda = (float) j/(float) mNbRbase* M_PI / 2.;
				*ptr = Vector3(mRadius * sin(lambda) + 0.5 * mHeight, 0., -mRadius * cos(lambda));
				ptr ++;		
				break;
			
			case BASE_FLAT:
				lambda = (float) j/(float) mNbRbase* M_PI / 2.;
				*ptr = Vector3(mEdgeRadius * sin(lambda) + 0.5 * mHeight, 0., -(mRadius - mEdgeRadius) - mEdgeRadius * cos(lambda));
				ptr ++;					
				break;
		}	
	}	
	if(mBase2 == BASE_FLAT) {
		*ptr = Vector3(0.5 * mHeight + mEdgeRadius, 0., 0.);
		ptr++;
	}	
		
	
	for(int i=1;i<mNbR;i++) {
		alpha = ((float) i/mNbR  - 1./2.)*2.*M_PI; 
		
		if(mBase1 == BASE_FLAT) {
			*ptr = Vector3(- 0.5 * mHeight - mEdgeRadius, 0., 0.);
			ptr++;
		}
		
		for(int j=-mNbRbase;j<0;j++) {			
			switch(mBase1) {
				case BASE_DOME:
					lambda = (float) j/(float) mNbRbase* M_PI / 2.;
					*ptr = Vector3(mRadius * sin(lambda) - 0.5 * mHeight, 
								   mRadius * sin(alpha) * cos(lambda), 
								   mRadius * cos(alpha) * cos(lambda));
					ptr ++;		
					break;
				
				case BASE_FLAT:
					lambda = (float) j/(float) mNbRbase* M_PI / 2.;
					*ptr = Vector3(mEdgeRadius * sin(lambda) - 0.5 * mHeight, 
								   ((mRadius - mEdgeRadius) + mEdgeRadius * cos(lambda)) * sin(alpha), 
								   ((mRadius - mEdgeRadius) + mEdgeRadius * cos(lambda)) * cos(alpha));
					ptr ++;					
					break;
			}
		}
		for(int j=0;j<=mNbH;j++) {
			*ptr = Vector3(((float) (j/mNbH) - 1./2.) * mHeight, mRadius * sin(alpha), mRadius * cos(alpha));
	    	ptr ++;
		}
		for(int j=1;j<=mNbRbase;j++) {			
			switch(mBase2) {
				case BASE_DOME:
					lambda = (float) j/(float) mNbRbase* M_PI / 2.;
					*ptr = Vector3(mRadius * sin(lambda) + 0.5 * mHeight, 
								   mRadius * sin(alpha) * cos(lambda), 
								   mRadius * cos(alpha) * cos(lambda));
					ptr ++;		
					break;
				
				case BASE_FLAT:
					lambda = (float) j/(float) mNbRbase* M_PI / 2.;
					*ptr = Vector3(mEdgeRadius * sin(lambda) + 0.5 * mHeight, 
								   ((mRadius - mEdgeRadius) + mEdgeRadius * cos(lambda)) * sin(alpha), 
								   ((mRadius - mEdgeRadius) + mEdgeRadius * cos(lambda)) * cos(alpha));
					ptr ++;				
					break;
			}
		}	
		
		if(mBase2 == BASE_FLAT) {
			*ptr = Vector3(0.5 * mHeight + mEdgeRadius, 0., 0.);
			ptr++;
		}		
	}
	
	
	if(mBase1 == BASE_FLAT) {
		*ptr = Vector3(- 0.5 * mHeight - mEdgeRadius, 0., 0.);
		ptr++;
	}
	for(int j=-mNbRbase;j<0;j++) {			
		switch(mBase1) {
			case BASE_DOME:
				lambda = (float) j/(float) mNbRbase* M_PI / 2.;
				*ptr = Vector3(mRadius * sin(lambda) - 0.5 * mHeight, 0., -mRadius * cos(lambda));
				ptr ++;		
				break;
			
			case BASE_FLAT:
				lambda = (float) j/(float) mNbRbase* M_PI / 2.;
				*ptr = Vector3(mEdgeRadius * sin(lambda) - 0.5 * mHeight, 0., -(mRadius - mEdgeRadius) - mEdgeRadius * cos(lambda));
				ptr ++;			
				break;
		}	
	}
	for(int j=0;j<=mNbH;j++) {		
		*ptr = Vector3(((float) (j/mNbH) - 1./2.) * mHeight, 0., -mRadius);
		ptr ++;
	}	
	for(int j=1;j<=mNbRbase;j++) {			
		switch(mBase2) {
			case BASE_DOME:
				lambda = (float) j/(float) mNbRbase* M_PI / 2.;
				*ptr = Vector3(mRadius * sin(lambda) + 0.5 * mHeight, 0., -mRadius * cos(lambda));
				ptr ++;		
				break;
			
			case BASE_FLAT:
				lambda = (float) j/(float) mNbRbase* M_PI / 2.;
				*ptr = Vector3(mEdgeRadius * sin(lambda) + 0.5 * mHeight, 0., -(mRadius - mEdgeRadius) - mEdgeRadius * cos(lambda));
				ptr ++;				
				break;
		}	
	}	
	if(mBase2 == BASE_FLAT) {
		*ptr = Vector3(0.5 * mHeight + mEdgeRadius, 0., 0.);
		ptr++;
	}	
}

void CylinderMesh::createColors(Color *ptr) {
	
}

void CylinderMesh::createNormals(Vector3 *ptr) {
	float alpha, lambda;
		
	if(mInside) {	
		
		for(int i=0;i<=mNbR;i++) {
			alpha = ((float) i/mNbR  - 1./2.)*2.*M_PI; 
			
			if(mBase1 == BASE_FLAT) {
				*ptr = Vector3(1., 0., 0.);
				ptr++;
			}	
			
			for(int j=-mNbRbase;j<0;j++) {			
				switch(mBase1) {
					case BASE_DOME:
						lambda = (float) j/(float) mNbRbase* M_PI / 2.;
						*ptr = Vector3(-sin(lambda), -sin(alpha) * cos(lambda), -cos(alpha) * cos(lambda));
						ptr ++;		
						break;
					
					case BASE_FLAT:
						lambda = (float) j/(float) mNbRbase* M_PI / 2.;
						*ptr = Vector3(-sin(lambda), -sin(alpha) * cos(lambda), -cos(alpha) * cos(lambda));
						ptr ++;			
						break;
				}
			}			
			
			for(int j=0;j<=mNbH;j++) {
				*ptr = Vector3(0., -sin(alpha), -cos(alpha));
				ptr ++;
			}
			
			for(int j=1;j<=mNbRbase;j++) {			
				switch(mBase2) {
					case BASE_DOME:
						lambda = (float) j/(float) mNbRbase* M_PI / 2.;
						*ptr = Vector3(-sin(lambda), -sin(alpha) * cos(lambda), -cos(alpha) * cos(lambda));
						ptr ++;		
						break;
					
					case BASE_FLAT:
						lambda = (float) j/(float) mNbRbase* M_PI / 2.;
						*ptr = Vector3(-sin(lambda), -sin(alpha) * cos(lambda), -cos(alpha) * cos(lambda));
						ptr ++;						
						break;
				}
			}	
			if(mBase2 == BASE_FLAT) {
				*ptr = Vector3(-1., 0., 0.);
				ptr++;
			}				
		}	
		
	} else {		
		
		for(int i=0;i<=mNbR;i++) {
			alpha = ((float) i/mNbR  - 1./2.)*2.*M_PI; 
			
			if(mBase1 == BASE_FLAT) {
				*ptr = Vector3(-1., 0., 0.);
				ptr++;
			}			
			
			for(int j=-mNbRbase;j<0;j++) {			
				switch(mBase1) {
					case BASE_DOME:
						lambda = (float) j/(float) mNbRbase* M_PI / 2.;
						*ptr = Vector3(sin(lambda), sin(alpha) * cos(lambda), cos(alpha) * cos(lambda));
						ptr ++;		
						break;
					
					case BASE_FLAT:
						lambda = (float) j/(float) mNbRbase* M_PI / 2.;
						*ptr = Vector3(sin(lambda), sin(alpha) * cos(lambda), cos(alpha) * cos(lambda));
						ptr ++;				
						break;
				}
			}		
			
			for(int j=0;j<=mNbH;j++) {
				*ptr = Vector3(0., sin(alpha), cos(alpha));
				ptr ++;
			}
			
			for(int j=1;j<=mNbRbase;j++) {			
				switch(mBase2) {
					case BASE_DOME:
						lambda = (float) j/(float) mNbRbase* M_PI / 2.;
						*ptr = Vector3(sin(lambda), sin(alpha) * cos(lambda), cos(alpha) * cos(lambda));
						ptr ++;		
						break;
					
					case BASE_FLAT:
						lambda = (float) j/(float) mNbRbase* M_PI / 2.;
						*ptr = Vector3(sin(lambda), sin(alpha) * cos(lambda), cos(alpha) * cos(lambda));
						ptr ++;						
						break;
				}
			}	

			if(mBase2 == BASE_FLAT) {
				*ptr = Vector3(1., 0., 0.);
				ptr++;
			}			
		}			
	}	
}

void CylinderMesh::createTexCoords(GLfloat *ptr) {
	
	//TODO
	
	/*float basesLenght = 0.;
	
	switch(mBase1) {
		case BASE_DOME:
			basesLenght += mRadius*M_PI/2.;
			break;
		
		case BASE_FLAT:
			basesLenght += mEdgeRadius*M_PI/2.+(mRadius-mEdgeRadius);
		
			break;
	}	
	switch(mBase2) {
		case BASE_DOME:
			basesLenght += mRadius*M_PI/2.;
			break;
		
		case BASE_FLAT:
			basesLenght += mEdgeRadius*M_PI/2.+(mRadius-mEdgeRadius);		
			break;
	}
	
	float texOffset = basesLenght/(mHeight+basesLenght);	
	
	for(int i=0;i<mNbR+1;i++){
		
		if(mBase1 == BASE_FLAT) {
			*ptr = 0.;
			ptr++;
			*ptr = 0.;
			ptr++;			
		}			
		for(int j=-mNbRbase;j<0;j++) {			
			switch(mBase1) {
				case BASE_DOME:
					*ptr = ((GLfloat) j/mNbRbase)*texOffset;
					ptr ++;
					*ptr = (GLfloat) i/mNbR;
					ptr ++;	
					break;
				
				case BASE_FLAT:
					*ptr = (mEdgeRadius / baseLenght) + ((GLfloat) j/mNbRbase)*texOffset * (baseLenght-mEdgeRadius) / baseLenght;
					ptr ++;
					*ptr = (GLfloat) i/mNbR;
					ptr ++;					
					break;
			}
		}	
		for(int j=0;j<mNbH+1;j++){
			*ptr = ((GLfloat) j/mNbH)*(1.-2.*texOffset) + texOffset;
	    	ptr ++;
			*ptr = (GLfloat) i/mNbR;
	    	ptr ++;		
		}
		for(int j=1;j<=mNbRbase;j++) {			
			switch(mBase2) {
				case BASE_DOME:
					*ptr = ((GLfloat) j/mNbRbase)*texOffset + (1.-texOffset);
					ptr ++;
					*ptr = (GLfloat) i/mNbR;
					ptr ++;	
					break;
				
				case BASE_FLAT:
					break;
			}
		}	
		
		if(mBase2 == BASE_FLAT) {
			*ptr = 1.;
			ptr++;
			*ptr = 1.;
			ptr++;			
		}		
	}	*/
}

void CylinderMesh::createFogCoords(GLfloat *ptr) {
	
}
		
void CylinderMesh::createTangents(Vector3 *ptr) {
	float alpha, lambda;
		
	for(int i=0;i<=mNbR;i++) {		
		alpha = ((float) i/mNbR  - 1./2.)*2.*M_PI; 
		
		if(mBase1 == BASE_FLAT) {
			*ptr = Vector3(0., sin((((float) i*2.-1.)/mNbR  - 1.)*M_PI), cos((((float) i*2.-1.)/mNbR  - 1.)*M_PI));
			ptr++;
		}
		
		for(int j=-mNbRbase;j<0;j++) {			
			switch(mBase1) {
				case BASE_DOME:
					lambda = (float) j/( 2. * (float) mNbRbase)* M_PI;					
					*ptr = Vector3(cos(lambda), -sin(alpha) * sin(lambda), -cos(alpha) * sin(lambda));
					ptr ++;		
					break;
				
				case BASE_FLAT:
					lambda = (float) j/( 2. * (float) mNbRbase)* M_PI;					
					*ptr = Vector3(cos(lambda), -sin(alpha) * sin(lambda), -cos(alpha) * sin(lambda));
					ptr ++;					
					break;
			}
		}		
		
		for(int j=0;j<=mNbH;j++) {
			*ptr = Vector3(1., 0., 0.);
			ptr ++;
		}
		
		for(int j=1;j<=mNbRbase;j++) {			
			switch(mBase2) {
				case BASE_DOME:
					lambda = (float) j/( 2. * (float) mNbRbase)* M_PI;					
					*ptr = Vector3(cos(lambda), -sin(alpha) * sin(lambda), -cos(alpha) * sin(lambda));
					ptr ++;		
					break;
				
				case BASE_FLAT:
					lambda = (float) j/( 2. * (float) mNbRbase)* M_PI;					
					*ptr = Vector3(cos(lambda), -sin(alpha) * sin(lambda), -cos(alpha) * sin(lambda));
					ptr ++;					
					break;
			}
		}	
		
		if(mBase2 == BASE_FLAT) {
			*ptr = Vector3(0., -sin(alpha), -cos(alpha));
			ptr++;
		}		
		
	}				
}
	
	
void CylinderMesh::createIndices(int numArray, GLuint *ptr) {

	int n = mNbH+1;
		
	switch(mBase1) {
		case BASE_DOME:
			n += mNbRbase;
			break;
		
		case BASE_FLAT:
			n += mNbRbase+1;		
			break;
	}	
	switch(mBase2) {
		case BASE_DOME:
			n += mNbRbase;
			break;
		
		case BASE_FLAT:
			n += mNbRbase+1;				
			break;
	}	
	
	if(numArray == mNbR) {
		if(mInside) {
			for(int j=0;j<n;j++){
				*ptr = numArray*n + j;
				ptr++;				
				*ptr = j;	
				ptr++;		
			}
		} else {
			for(int j=0;j<n;j++){
				*ptr = j;	
				ptr++;
				*ptr = numArray*n + j;
				ptr++;
			}
		}		
		
	} else {
		if(mInside) {
			for(int j=0;j<n;j++){
				*ptr = numArray*n + j;
				ptr++;				
				*ptr = (numArray +1)*n + j;	
				ptr++;		
			}
		} else {
			for(int j=0;j<n;j++){
				*ptr = (numArray +1)*n + j;	
				ptr++;
				*ptr = numArray*n + j;
				ptr++;
			}
		}
	}	
}
	
